"""
The main module for manipulating dates that are not ``datetime`` datatypes.
Consisting of a main class object, ``YMD``, for managing a date as an integer
and several functions for further functionality, such as adding or calculating differences.
All of these are hopefully self explanitory.
"""
# -----------------------------------------------

from ._tools import *

# -----------------------------------------------
# End.
